from __future__ import annotations

import statistics as stats
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, Optional, TypeVar

import numpy as np
from numpy.typing import NDArray

try:
    from rtamt import (
        LTLPastifyException,
        Semantics,
        STLDenseTimeSpecification,
        STLDiscreteTimeSpecification,
    )
except ImportError:
    _has_rtamt = False
else:
    _has_rtamt = True

from .core import Specification, SpecificationError

try:
    from .parser import parse
except:
    _can_parse = False
else:
    _can_parse = True


StateT = TypeVar("StateT")
PredicateName = str
ColumnT = int
PredicateColumnMap = Dict[PredicateName, ColumnT]


def _valid_trajectories_array(trajectories: NDArray[Any], row_len: int) -> Optional[str]:
    if not trajectories.ndim == 2:
        return "trajectories array must be 2-dimensional"

    if not trajectories.shape[1] == row_len:
        return f"trajectory array must have {row_len} columns"

    return None


class StlSpecification(Specification[StateT], ABC):
    @abstractmethod
    def __init__(self, requirement: str, column_map: PredicateColumnMap):
        ...


class TLTK(StlSpecification[NDArray[np.float_]]):
    """STL logic specification that uses TLTK to compute robustness values.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, column_map: PredicateColumnMap):
        if not _can_parse:
            raise RuntimeError(
                "TLTK specifications require parsing functionality. Please refer to the documentation for how to enable parsing."
            )

        tltk_obj = parse(phi, list(column_map.keys()))

        if tltk_obj is None:
            raise SpecificationError("could not parse formula")

        self.tltk_obj = tltk_obj
        self.column_map = column_map

    def evaluate(self, states: NDArray[np.float_], times: NDArray[np.float_]) -> float:
        trajectories_err = _valid_trajectories_array(states, times.size)

        if trajectories_err is not None:
            raise SpecificationError(trajectories_err)

        map_items = self.column_map.items()
        traces = {name: np.array(states[column], dtype=np.float64) for name, column in map_items}
        timestamps = np.array(times, dtype=np.float32)

        self.tltk_obj.reset()
        self.tltk_obj.eval_interval(traces, timestamps)

        return self.tltk_obj.robustness


def _step_widths(times: NDArray[np.float_]) -> Iterable[float]:
    """Compute the distance between adjacent elements."""

    for i in range(len(times) - 2):
        yield abs(times[i] - times[i + 1])


class RTAMTDiscrete(StlSpecification[NDArray[np.float_]]):
    """STL logic specification that uses RTAMT discrete-time semantics to compute robustness.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, column_map: PredicateColumnMap):
        if not _has_rtamt:
            raise RuntimeError("RTAMT must be installed to use RTAMTDiscrete specification")

        if "time" in column_map:
            raise SpecificationError("'time' cannot be used as a predicate name for RTAMT")

        self.rtamt_obj = STLDiscreteTimeSpecification(Semantics.STANDARD)

        self.rtamt_obj.spec = phi
        self.column_map = column_map

        for name in column_map:
            self.rtamt_obj.declare_var(name, "float")

    def evaluate(self, states: NDArray[np.float_], times: NDArray[np.float_]) -> float:
        trajectories_err = _valid_trajectories_array(states, times.size)

        if trajectories_err is not None:
            raise SpecificationError(trajectories_err)

        self.rtamt_obj.reset()

        period = stats.mean(_step_widths(times))
        self.rtamt_obj.set_sampling_period(round(period, 2), "s", 0.1)

        # parse AFTER declaring variables and setting sampling period
        self.rtamt_obj.parse()

        try:
            self.rtamt_obj.pastify()
        except LTLPastifyException:
            pass

        traces = {"time": times.tolist()}

        for name, column in self.column_map.items():
            traces[name] = states[column].tolist()

        # traces: Dict['time': timestamps, 'variable'(s): trajectories]
        robustness = self.rtamt_obj.evaluate(traces)

        return robustness[-1][1]


class RTAMTDense(StlSpecification[NDArray[np.float_]]):
    """STL logic specification that uses RTAMT dense-time semantics to compute robustness.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, column_map: PredicateColumnMap):
        if not _has_rtamt:
            raise RuntimeError("RTAMT must be installed to use RTAMTDense specification")

        self.rtamt_obj = STLDenseTimeSpecification(Semantics.STANDARD)
        self.column_map = column_map
        self.rtamt_obj.spec = phi

        for name in column_map:
            self.rtamt_obj.declare_var(name, "float")

    def evaluate(self, states: NDArray[np.float_], times: NDArray[np.float_]) -> float:
        trajectories_err = _valid_trajectories_array(states, times.size)

        if trajectories_err is not None:
            raise SpecificationError(trajectories_err)

        self.rtamt_obj.reset()

        # parse AFTER declaring variables
        self.rtamt_obj.parse()

        try:
            self.rtamt_obj.pastify()
        except LTLPastifyException:
            pass

        map_items = self.column_map.items()
        traces = [
            (name, np.array([times, states[column]]).T.tolist()) for name, column in map_items
        ]

        robustness = self.rtamt_obj.evaluate(*traces)

        return robustness[-1][1]



from __future__ import annotations

import statistics as stats
from typing import Generic, Iterable, Optional, Sequence, TypeVar

from attr import frozen

from .cost import Evaluation
from .interval import Interval

RT = TypeVar("RT")
ET = TypeVar("ET")


@frozen(slots=True)
class TimeStats:
    """Data class that represents the standard statistics of a set of durations."""

    durations: Iterable[float]

    @property
    def total_duration(self) -> float:
        """The sum of all durations."""

        return sum(self.durations)

    @property
    def avg_duration(self) -> float:
        """The average of all durations."""

        return stats.mean(self.durations)

    @property
    def max_duration(self) -> float:
        """The maximum duration."""

        return max(self.durations)

    @property
    def min_duration(self) -> float:
        """The maximum duration."""

        return min(self.durations)


@frozen(slots=True)
class Run(Generic[RT, ET]):
    """Data class that represents one run of an optimizer.

    Attributes:
        result: The value returned by the optimizer
        history: List of Evaluation instances representing every cost function evaluation
        duration: Time spent by the optimizer
    """

    result: RT
    history: Sequence[Evaluation[ET]]
    duration: float
    seed: int

    @property
    def worst_eval(self) -> Evaluation[ET]:
        """The evaluation with the highest cost (furthest from falsifying)."""

        return max(self.history, key=lambda e: e.cost)

    @property
    def best_eval(self) -> Evaluation[ET]:
        """The evaluation with the lowest cost (closest to falsification)."""

        return min(self.history, key=lambda e: e.cost)

    @property
    def fastest_eval(self) -> Evaluation[ET]:
        """The evaluation with the lowest total duration."""

        return min(self.history, key=lambda e: e.timing.total)

    @property
    def slowest_eval(self) -> Evaluation[ET]:
        """Evaluation with the longest total duration."""

        return max(self.history, key=lambda e: e.timing.total)

    @property
    def model_timing(self) -> TimeStats:
        """Time statistics for the model."""

        return TimeStats(iteration.timing.model for iteration in self.history)

    @property
    def specification_timing(self) -> TimeStats:
        """Time statistics of the specification."""

        return TimeStats(iteration.timing.specification for iteration in self.history)


@frozen(slots=True)
class Result(Generic[RT, ET]):
    """Data class that represents a set of successful runs of the optimizer.

    Attributes:
        runs: List of Run instances representing each optimization attempt
        options: Configuration class used to control Model, Specification and Optimizer behaviors
                 for each run
    """

    runs: Sequence[Run[RT, ET]]
    interval: Interval
    seed: int
    processes: Optional[int]

    @property
    def worst_run(self) -> Run[RT, ET]:
        return max(self.runs, key=lambda r: r.worst_eval.cost)

    @property
    def best_run(self) -> Run[RT, ET]:
        return min(self.runs, key=lambda r: r.best_eval.cost)



from __future__ import annotations

from typing import Dict, Optional, Sequence, Union

import tltk_mtl as mtl
from antlr4.CommonTokenStream import CommonTokenStream
from antlr4.InputStream import InputStream

from .stlLexer import stlLexer as Lexer
from .stlParser import stlParser as Parser
from .stlParserVisitor import stlParserVisitor as Visitor

PredicateNameSeq = Sequence[str]
PredicateDict = Dict[str, mtl.Predicate]
Predicates = Union[PredicateNameSeq, PredicateDict]
TltkObject = Union[
    mtl.And,
    mtl.Finally,
    mtl.Global,
    mtl.Implication,
    mtl.Next,
    mtl.Not,
    mtl.Or,
    mtl.Predicate,
    mtl.Until,
]


def parse(formula: str, predicates: Predicates, mode: str = "cpu") -> Optional[TltkObject]:
    """TLTk parser parses a specification requirement into an equivalent TLTk structure

    Attributes:
        formula: The formal specification requirement
        predicates: The set of Predicate(s) used in the requirement
        mode: The TLTk computation mode
    """
    input_stream = InputStream(formula)

    lexer = Lexer(input_stream)
    stream = CommonTokenStream(lexer)

    parser = Parser(stream)
    tree = parser.stlSpecification()
    visitor = Visitor(lexer, predicates, mode)

    return visitor.visit(tree)  # type: ignore



from .interval import Interval
from .model import Failure, Model, ModelData, ModelError
from .optimizer import ObjectiveFn, Optimizer
from .result import Result
from .sample import Sample
from .scenario import Scenario
from .signal import Signal, SignalFactory
from .specification import Specification, SpecificationError

__all__ = [
    "Interval",
    "Model",
    "ModelData",
    "Failure",
    "ModelError",
    "Optimizer",
    "ObjectiveFn",
    "Result",
    "Sample",
    "Scenario",
    "Signal",
    "SignalFactory",
    "Specification",
    "SpecificationError",
]



from __future__ import annotations

from typing import Iterable, List, TypeVar, cast

import numpy as np

from .core.cost import SignalParameters, SpecificationOrFactory, decompose_sample
from .core.interval import Interval
from .core.model import Model, ModelResult
from .core.optimizer import Optimizer
from .core.result import Result
from .core.sample import Sample
from .core.scenario import Scenario
from .options import Options, SignalOptions

StateT = TypeVar("StateT")
ResultT = TypeVar("ResultT")
ExtraT = TypeVar("ExtraT")


def _signal_parameters(options: Options, start_index: int) -> List[SignalParameters]:
    t_span = options.interval
    signal_parameters = []

    for signal in options.signals:
        n_control_points = len(signal.control_points)
        stop_index = start_index + n_control_points
        values_range = slice(start_index, stop_index, 1)

        if signal.signal_times is None:
            times_array = np.linspace(
                t_span.lower, t_span.upper, num=n_control_points, dtype=np.float64
            )
            signal_times = cast(List[float], times_array.tolist())
        else:
            signal_times = signal.signal_times

        parameters = SignalParameters(values_range, signal_times, signal.factory)
        signal_parameters.append(parameters)
        start_index = stop_index

    return signal_parameters


def _signal_bounds(signals: Iterable[SignalOptions]) -> tuple[Interval, ...]:
    return sum((signal.control_points for signal in signals), ())


def staliro(
    model: Model[StateT, ExtraT],
    specification: SpecificationOrFactory[StateT],
    optimizer: Optimizer[ResultT],
    options: Options,
) -> Result[ResultT, ExtraT]:
    """Search for falsifying inputs to the provided system.

    Using the optimizer, search the input space defined in the options for cases which falsify the
    provided specification represented by the formula phi and the predicates.

    Args:
        model: The model of the system being tested.
        specification: The requirement for the system.
        optimizer: The optimizer to use to search the sample space.
        options: General options to manipulate the behavior of the search.

    Returns:
        An object containing the result of each optimization attempt.
    """
    static_params_end = len(options.static_parameters)
    signal_parameters = _signal_parameters(options, static_params_end)
    signal_bounds = _signal_bounds(options.signals)
    bounds = options.static_parameters + signal_bounds

    scenario: Scenario[StateT, ResultT, ExtraT] = Scenario(
        model=model,
        specification=specification,
        runs=options.runs,
        iterations=options.iterations,
        seed=options.seed,
        processes=options.process_count,
        bounds=bounds,
        interval=options.interval,
        static_parameter_range=slice(0, static_params_end, 1),
        signal_parameters=signal_parameters,
    )

    return scenario.run(optimizer)


def simulate_model(
    model: Model[StateT, ExtraT], options: Options, sample: Sample
) -> ModelResult[StateT, ExtraT]:
    static_range = slice(0, len(options.static_parameters))
    signal_params = _signal_parameters(options, static_range.stop)
    static_inputs, signals = decompose_sample(sample, static_range, signal_params)

    return model.simulate(static_inputs, signals, options.interval)



from __future__ import annotations

import logging
import math
import time
from concurrent.futures import ProcessPoolExecutor
from typing import Any, Callable, Generic, Iterable, Iterator, Sequence, Tuple, TypeVar, Union, cast

from attr import field, frozen
from attr.validators import instance_of

from .interval import Interval
from .model import Failure, Model, ModelData, ModelError, ModelResult
from .optimizer import ObjectiveFn
from .sample import Sample
from .signal import Signal, SignalFactory
from .specification import Specification, SpecificationError

StateT = TypeVar("StateT")
ExtraT = TypeVar("ExtraT")

SpecificationFactory = Callable[[Sample], Specification[StateT]]
SpecificationOrFactory = Union[Specification[StateT], SpecificationFactory[StateT]]

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _slice_length(s: slice) -> int:
    return cast(int, s.stop - s.start // s.step)


@frozen()
class TimingData:
    """Storage class for execution durations of different PSY-TaLiRo components.

    The durations stored in this class are for a single evaluation.

    Attributes:
        model: Run time of model component
        specification: Run time of specification component
    """

    model: float
    specification: float

    @property
    def total(self) -> float:
        """The total duration of all components."""

        return self.model + self.specification


@frozen()
class Evaluation(Generic[ExtraT]):
    """The result of applying the cost function to a sample.

    Attributes:
        cost: The result of using a specification to analyze the output of a model
        sample: The sample provided to the model
        extra: Additional data returned by the model
        timing: Execution durations of each component of the cost function
    """

    cost: float
    sample: Sample
    extra: ExtraT
    timing: TimingData


class EvaluationError(Exception):
    pass


T = TypeVar("T")


def _time(func: Callable[[], T]) -> Tuple[float, T]:
    start_time = time.perf_counter()
    result = func()
    stop_time = time.perf_counter()
    duration = stop_time - start_time

    return duration, result


def _mksignal(sample: Sample, params: SignalParameters) -> Signal:
    signal_values = sample[params.values_range]

    if len(signal_values) != _slice_length(params.values_range):
        raise EvaluationError(
            f"cannot satisfy signal values range {params.values_range} with sample {sample}"
        )

    signal = params.factory(params.times, signal_values)

    if not isinstance(signal, Signal):
        raise EvaluationError(
            f"unknown type {type(signal)} returned from signal factory {params.factory.__class__}"
        )

    return signal


def _result_cost(specification: Specification[StateT], result: ModelResult[StateT, Any]) -> float:
    if isinstance(result, ModelData):
        return specification.evaluate(result.states, result.times)
    elif isinstance(result, Failure):
        return -math.inf

    raise ModelError("unsupported type returned from model")


@frozen(auto_attribs=False, slots=True)
class SignalParameters:
    values_range: slice = field()
    times: Sequence[float] = field()
    factory: SignalFactory = field()


def decompose_sample(
    sample: Sample, static_range: slice, params_seq: Sequence[SignalParameters]
) -> Tuple[Sequence[float], Sequence[Signal]]:
    static_parameters = sample[static_range]
    signals = [_mksignal(sample, params) for params in params_seq]

    return static_parameters, signals


@frozen()
class Thunk(Generic[StateT, ExtraT]):
    """Class which represents the deferred evaluation of the cost function.

    A Thunk contains all the necessary information to produce an Evaluation, but does not execute
    the computation until the evaluate method is called (lazy value).

    Attributes:
        sample: The sample generated by the optimizer that is provided to the model
        model: Model instance which represents the system
        specification: Specification instance which represents the requirement
        options: Configuration object that controls the behavior of the Model and Specification
    """

    sample: Sample = field(validator=instance_of(Sample))
    model: Model[StateT, ExtraT]
    _specification: SpecificationOrFactory[StateT]
    interval: Interval
    static_parameter_range: slice
    signal_parameters: Sequence[SignalParameters]

    @property
    def specification(self) -> Specification[StateT]:
        if not callable(self._specification):
            return self._specification

        specification = self._specification(self.sample)

        if not isinstance(specification, Specification):
            raise SpecificationError(f"invalid type {type(specification)} given as specification")

        return specification

    def evaluate(self) -> Evaluation[ExtraT]:
        """Evaluate the sample using the specification and model.

        The computation is the following pipeline:
            sample -> model.simulate(sample) -> specification.evaluate(model_result) -> cost

        Returns:
            An Evaluation instance representing the result of the computation pipeline.
        """

        static_inputs, signals = decompose_sample(
            self.sample,
            self.static_parameter_range,
            self.signal_parameters,
        )
        simulate = lambda: self.model.simulate(static_inputs, signals, self.interval)
        model_duration, model_result = _time(simulate)

        compute_cost = lambda: _result_cost(self.specification, model_result)
        cost_duration, cost = _time(compute_cost)
        timing_data = TimingData(model_duration, cost_duration)

        return Evaluation(cost, self.sample, model_result.extra, timing_data)


@frozen()
class ThunkGenerator(Generic[StateT, ExtraT], Iterable[Thunk[StateT, ExtraT]]):
    """Generate Thunk instances from a collection of samples."""

    samples: Iterable[Sample]
    model: Model[StateT, ExtraT]
    specification: SpecificationOrFactory[StateT]
    interval: Interval
    static_parameter_range: slice
    signal_parameters: Sequence[SignalParameters]

    def __iter__(self) -> Iterator[Thunk[StateT, ExtraT]]:
        for sample in self.samples:
            yield Thunk(
                sample,
                self.model,
                self.specification,
                self.interval,
                self.static_parameter_range,
                self.signal_parameters,
            )


def _evaluate(thunk: Thunk[Any, ExtraT]) -> Evaluation[ExtraT]:
    return thunk.evaluate()


@frozen()
class CostFn(ObjectiveFn, Generic[StateT, ExtraT]):
    """Class which represents the composition of a Model and Specification.

    A Model is responsible for modeling the system and returning a trajectory given a sample, and a
    Specification is responsible for evaluating a system trajectory and returning a value which
    indicates its "goodness". Composing these two components produces a component which can produce
    "goodness" values given a sample.

    Attributes:
        model: Model instance which represents a system
        specification: Specification instance or factory which represents the requirement
        options: Configuration object that controls the behavior of the Model and Specification
        history: List of all evaluations produced by this instance
    """

    model: Model[StateT, ExtraT]
    specification: SpecificationOrFactory[StateT]
    interval: Interval
    static_parameter_range: slice
    signal_parameters: Sequence[SignalParameters]
    history: list[Evaluation[ExtraT]] = field(init=False, factory=list)

    def eval_sample(self, sample: Sample) -> float:
        """Compute the cost of a single sample.

        Args:
            sample: Sample to evaluate

        Returns:
            The sample cost
        """

        logger.debug(f"Evaluating sample {sample}")

        thunk = Thunk(
            sample,
            self.model,
            self.specification,
            self.interval,
            self.static_parameter_range,
            self.signal_parameters,
        )
        evaluation = _evaluate(thunk)

        self.history.append(evaluation)

        return evaluation.cost

    def eval_samples(self, samples: Sequence[Sample]) -> list[float]:
        """Compute the cost of multiple samples sequentially.

        Args:
            samples: Samples to evaluate

        Returns:
            The cost for each sample
        """

        logger.debug(f"Evaluating samples {samples}")

        thunks = ThunkGenerator(
            samples,
            self.model,
            self.specification,
            self.interval,
            self.static_parameter_range,
            self.signal_parameters,
        )
        evaluations = [_evaluate(thunk) for thunk in thunks]

        self.history.extend(evaluations)

        return [evaluation.cost for evaluation in evaluations]

    def eval_samples_parallel(self, samples: Sequence[Sample], processes: int) -> list[float]:
        """Compute the cost of multiple samples in parallel.

        Samples are evaluated row-wise, so each row is considered a different sample.

        Args:
            samples: The samples to evaluate
            processes: Number of processes to use to evaluate the samples

        Returns:
            The cost for each sample.
        """

        logger.debug(f"Evaluating samples {samples} with {processes} processes")

        thunks = ThunkGenerator(
            samples,
            self.model,
            self.specification,
            self.interval,
            self.static_parameter_range,
            self.signal_parameters,
        )

        with ProcessPoolExecutor(max_workers=processes) as executor:
            futures: Iterable[Evaluation[ExtraT]] = executor.map(_evaluate, thunks)
            evaluations = list(futures)

        self.history.extend(evaluations)

        return [evaluation.cost for evaluation in evaluations]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['staliro', 'staliro.core', 'staliro.parser']

package_data = \
{'': ['*'], 'staliro.parser': ['grammar/*']}

install_requires = \
['antlr4-python3-runtime>=4.5.3,<4.6.0',
 'attrs>=21.0.0,<22.0.0',
 'numpy>=1.21.5,<2.0.0',
 'scipy>=1.6.2,<2.0.0',
 'typing-extensions>=4.2.0,<5.0.0']

setup_kwargs = {
    'name': 'psy-taliro',
    'version': '1.0.0b5',
    'description': 'System-level verification library using STL',
    'long_description': None,
    'author': 'Quinn Thibeault',
    'author_email': 'qthibeau@asu.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, Sequence, TypeVar, Union, overload

import numpy as np
from attr import Attribute, field, frozen
from numpy.typing import NDArray

from .interval import Interval
from .signal import Signal

Times = NDArray[np.float_]


def _times_validator(obj: Any, attr: Attribute[Any], times: Times) -> None:
    if not isinstance(times, np.ndarray):
        raise TypeError("timestamps must be provided as a NDArray")

    if not np.issubdtype(times.dtype, np.floating):
        raise TypeError("timestamp values must be floating point values")

    if times.ndim != 1:
        raise ValueError("timestamps must be 1-dimensional")


StateT = TypeVar("StateT")
ExtraT = TypeVar("ExtraT")


@frozen(init=False)
class ModelData(Generic[StateT, ExtraT]):
    """Representation of the state of a system over time.

    Attributes:
        times: The timestamps corresponding to each state of the system
        extra: User-defined data related to the particular system execution
    """

    states: StateT
    times: Times = field(validator=_times_validator, converter=np.array)
    extra: ExtraT

    @overload
    def __init__(self: ModelData[StateT, None], states: StateT, timestamps: Any):
        ...

    @overload
    def __init__(self, states: StateT, timestamps: Any, extra: ExtraT):
        ...

    def __init__(self, states: StateT, timestamps: Any, extra: ExtraT = None):
        self.__attrs_init__(states, timestamps, extra)  # type: ignore


@frozen(init=False)
class Failure(Generic[ExtraT]):
    """Representation of a system failure that should be interpreted as a falsification.

    Attributes:
        extra: User-defined data related to the particular system execution
    """

    extra: ExtraT = field()

    @overload
    def __init__(self: Failure[None]):
        ...

    @overload
    def __init__(self, extra: ExtraT):
        ...

    def __init__(self, extra: ExtraT = None):
        self.__attrs_init__(extra)  # type: ignore


StaticInput = Sequence[float]
Signals = Sequence[Signal]
ModelResult = Union[ModelData[StateT, ExtraT], Failure[ExtraT]]


class Model(Generic[StateT, ExtraT], ABC):
    """A model is a representation of a system under test (SUT).

    The model defines a simulate method that contains the logic necessary to simulate the system
    given a set of inputs.
    """

    @abstractmethod
    def simulate(
        self, static: StaticInput, signals: Signals, interval: Interval
    ) -> ModelResult[StateT, ExtraT]:
        """Simulate the model using the given inputs.

        This method contains the logic responsible for simulating the model with the given inputs.
        The result of this method should be a ModelResult instance containing a set of timestamps
        and state values representing the execution of the system over time, or a Failure instance
        that serves as an indicator of a system failure that should be interpreted as a
        falsification of the system requirement.

        Arguments:
            static: The static inputs to the system
            signals: time-varying inputs to the system represented as interpolated functions

        Returns:
            An instance of ModelData indicating a successful simulation, or a Failure if an error is
            encountered that should result in a falsification.
        """
        ...


class ModelError(Exception):
    pass
